/* 
 * Copyright (C) 2000 the xine project
 * 
 * This file is part of xine, a unix video player.
 * 
 * xine is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * xine is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 * $Id: demux_mpeg.c,v 1.14 2000/12/12 18:45:20 guenter Exp $
 *
 * demultiplexer for mpeg 1/2 program streams
 * reads streams of variable blocksizes
 *
 * currently only used for mpeg-1-files
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <pthread.h>
#include <string.h>

#include "monitor.h"
#include "xine.h"
#include "demux.h"

extern uint32_t xine_debug;

typedef struct _demux_mpeg_globals {
  fifo_buffer_t       *mBufAudio;
  fifo_buffer_t       *mBufVideo;

  input_plugin_t      *mInput;

  pthread_t            mThread;

  unsigned char        mDummySpace[8192];

  int                  mnAudioChannel;
  int                  mnStatus;
} demux_mpeg_globals_t ;

static demux_mpeg_globals_t gDemuxMpeg;

static uint32_t read_bytes (int n) {
  
  uint32_t res;
  uint32_t i;
  unsigned char buf[6];

  buf[4]=0;

  i = gDemuxMpeg.mInput->read (buf, n);

  if (i != n) {
    gDemuxMpeg.mnStatus = DEMUX_FINISHED;
    xprintf (VERBOSE|DEMUX, "Unexpected end of stream\n");
  }
  

  switch (n)  {
  case 1:
    res = buf[0];
    break;
  case 2:
    res = (buf[0]<<8) | buf[1];
    break;
  case 3:
    res = (buf[0]<<16) | (buf[1]<<8) | buf[2];
    break;
  case 4:
    res = (buf[2]<<8) | buf[3] | (buf[1]<<16) | (buf[0] << 24);
    break;
  default:
    fprintf (stderr,
	     "How how - something wrong in wonderland demux:read_bytes (%d)\n",
	     n);
    exit (1);
  }

  return res;
}

void parse_mpeg2_packet (int nID) {
  int            nLen, i, nTrack;
  uint32_t       buf, flags, header_len, pts;
  buf_element_t *pBuf;

  nLen = read_bytes(2);

  xprintf (VERBOSE|DEMUX|MPEG, "  mpeg2 packet (len=%d",nLen); 

  if (nID==0xbd) {

    xprintf (VERBOSE|DEMUX|AC3, ",ac3");

    buf = read_bytes(1);
    flags = read_bytes(1);
    header_len = read_bytes(1);

    nLen -= header_len + 3;

    pts=0;

    if ((flags & 0x80) == 0x80) {

      buf = read_bytes(1);
      pts = (buf & 0x0e) << 29 ;
      buf = read_bytes(2); 
      pts |= (buf & 0xFFFE) << 14;
      buf = read_bytes(2); 
      pts |= (buf & 0xFFFE) >> 1;

      xprintf (VERBOSE|DEMUX|VPTS, ", pts=%d",pts);

      header_len -= 5 ;
    }

    /* read rest of header */
    i = gDemuxMpeg.mInput->read (gDemuxMpeg.mDummySpace, header_len+4);

    nTrack = gDemuxMpeg.mDummySpace[0] & 0x0F ;

    xprintf (VERBOSE|DEMUX, ", track=%02x",nTrack);

    /* contents */

    if (nTrack == gDemuxMpeg.mnAudioChannel) {

      pBuf = buffer_pool_alloc ();

      i = (*gDemuxMpeg.mInput->read) (pBuf->pMem, nLen-4);

      pBuf->pContent  = pBuf->pMem;
      pBuf->nSize     = nLen-4;
      pBuf->nType     = BUF_AC3AUDIO;
      pBuf->nPTS      = pts;
      pBuf->nDTS      = 0 ; /* FIXME */
      pBuf->nInputPos = gDemuxMpeg.mInput->seek (0, SEEK_CUR);

      fifo_buffer_put (gDemuxMpeg.mBufAudio, pBuf);

    } else {
      i = gDemuxMpeg.mInput->read (gDemuxMpeg.mDummySpace, nLen-4); 

      /* (*gDemuxMpeg.mInput->seek) (nLen-4, SEEK_CUR); */

    }

  } else if ((nID & 0xe0) == 0xc0) {

    xprintf (VERBOSE|DEMUX|AUDIO, ", audio #%d", nID & 0x1f);


    buf = read_bytes(1);
    flags = read_bytes(1);
    header_len = read_bytes(1);

    nLen -= header_len + 3;

    pts = 0;

    if ((flags & 0x80) == 0x80) {

      buf = read_bytes(1);
      pts = (buf & 0x0e) << 29 ;
      buf = read_bytes(2); 
      pts |= (buf & 0xFFFE) << 14;
      buf = read_bytes(2); 
      pts |= (buf & 0xFFFE) >> 1;

      xprintf (VERBOSE|DEMUX|VPTS, ", pts=%d",pts);

      header_len -= 5 ;
    }

    /* read rest of header */
    i = gDemuxMpeg.mInput->read (gDemuxMpeg.mDummySpace, header_len);


    pBuf = buffer_pool_alloc ();

    i = gDemuxMpeg.mInput->read (pBuf->pMem, nLen);

    pBuf->pContent  = pBuf->pMem;
    pBuf->nSize     = nLen;
    pBuf->nType     = BUF_MPEGAUDIO;
    pBuf->nPTS      = pts;
    pBuf->nDTS      = 0;   /* FIXME */
    pBuf->nInputPos = gDemuxMpeg.mInput->seek (0, SEEK_CUR);

    fifo_buffer_put (gDemuxMpeg.mBufAudio, pBuf);

  } else if ((nID >= 0xbc) && ((nID & 0xf0) == 0xe0)) {

    xprintf (VERBOSE|DEMUX|VIDEO, ",video");

    buf = read_bytes(1);
    flags = read_bytes(1);
    header_len = read_bytes(1);

    nLen -= header_len + 3;

    pts = 0;

    if ((flags & 0x80) == 0x80) {

      buf = read_bytes(1);
      pts = (buf & 0x0e) << 29 ;
      buf = read_bytes(2); 
      pts |= (buf & 0xFFFE) << 14;
      buf = read_bytes(2); 
      pts |= (buf & 0xFFFE) >> 1;

      xprintf (VERBOSE|DEMUX|VPTS, ", pts=%d",pts);

      header_len -= 5 ;
    }

    /* read rest of header */
    i = gDemuxMpeg.mInput->read (gDemuxMpeg.mDummySpace, header_len);

    /* contents */

    pBuf = buffer_pool_alloc ();

    i = (*gDemuxMpeg.mInput->read) (pBuf->pMem, nLen);

    pBuf->pContent = pBuf->pMem;
    pBuf->nSize = nLen;
    pBuf->nType = BUF_MPEG2VIDEO;
    pBuf->nPTS  = pts;
    pBuf->nDTS  = 0;

    fifo_buffer_put (gDemuxMpeg.mBufVideo, pBuf);

  } else {
    xprintf (VERBOSE|DEMUX, ",unknown stream - skipped");

    i = gDemuxMpeg.mInput->read (gDemuxMpeg.mDummySpace, nLen); 
    /* (*gDemuxMpeg.mInput->seek) (nLen,SEEK_CUR); */
  }

  xprintf (VERBOSE|DEMUX, ")\n");

}

void parse_mpeg1_packet (int nID) 
{
  int             nLen;
  uint32_t        buf;
  int             i;
  int             pts;
  buf_element_t  *pBuf;

  xprintf (VERBOSE|DEMUX, "  packet (");

  nLen = read_bytes(2);

  xprintf (VERBOSE|DEMUX, "len=%d",nLen);

  pts=0;

  if (nID != 0xbf) {

    buf = read_bytes(1); nLen--;

    while ((buf & 0x80) == 0x80)   {

      if (gDemuxMpeg.mnStatus != DEMUX_OK)
	return;

      /* stuffing bytes */
      buf = read_bytes(1); nLen--;
    }

    if ((buf & 0xC0) == 0x40) {

      if (gDemuxMpeg.mnStatus != DEMUX_OK)
	return;

      /* buffer_scale, buffer size */
      buf = read_bytes(1); nLen--;
      buf = read_bytes(1); nLen--;
    }

    if ((buf & 0xF0) == 0x20) {

      if (gDemuxMpeg.mnStatus != DEMUX_OK)
	return;

      pts = (buf & 0xe) << 29 ;
      buf = read_bytes(2); nLen -= 2;

      pts |= (buf & 0xFFFE) << 14;

      buf = read_bytes(2); nLen -= 2;
      pts |= (buf & 0xFFFE) >> 1;

      xprintf (VERBOSE|DEMUX|VPTS, ", pts=%d",pts);

      /* pts = 0; */

    } else if ((buf & 0xF0) == 0x30) {

      if (gDemuxMpeg.mnStatus != DEMUX_OK)
	return;

      pts = (buf & 0x0e) << 29 ;
      buf = read_bytes(2); nLen -= 2;
      
      pts |= (buf & 0xFFFE) << 14;

      buf = read_bytes(2); nLen -= 2;

      pts |= (buf & 0xFFFE) >> 1;

/*       printf ("pts2=%d\n",pts); */
      xprintf (VERBOSE|DEMUX|VPTS, ", pts2=%d",pts);

      /* Decoding Time Stamp */
      buf = read_bytes(3); nLen -= 3;
      buf = read_bytes(2); nLen -= 2;
    } else {
      xprintf (VERBOSE|DEMUX, ", buf = %02x",buf);
      if (buf != 0x0f)
	xprintf (VERBOSE|DEMUX, " ERROR buf (%02x) != 0x0F ",buf);
    }

  }

  if ((nID & 0xe0) == 0xc0) {
    xprintf (VERBOSE|DEMUX|AUDIO, ", audio #%d", nID & 0x1f);

    pBuf = buffer_pool_alloc ();

    i = (*gDemuxMpeg.mInput->read) (pBuf->pMem, nLen);

    pBuf->pContent  = pBuf->pMem;
    pBuf->nSize     = nLen;
    pBuf->nType     = BUF_MPEGAUDIO;
    pBuf->nPTS      = pts;
    pBuf->nDTS      = 0;   /* FIXME */
    pBuf->nInputPos = gDemuxMpeg.mInput->seek (0, SEEK_CUR);

    fifo_buffer_put (gDemuxMpeg.mBufAudio, pBuf);

  } else if ((nID & 0xf0) == 0xe0) {

    xprintf (VERBOSE|DEMUX|VIDEO, ", video #%d", nID & 0x0f);

    pBuf = buffer_pool_alloc ();

    i = (*gDemuxMpeg.mInput->read) (pBuf->pMem, nLen);

    pBuf->pContent = pBuf->pMem;
    pBuf->nSize = nLen;
    pBuf->nType = BUF_MPEG1VIDEO;
    pBuf->nPTS  = pts;
    pBuf->nDTS  = 0; /* FIXME */

    fifo_buffer_put (gDemuxMpeg.mBufVideo, pBuf);

  } else if (nID == 0xbd) {
    xprintf (VERBOSE|DEMUX|AC3, ", ac3");
    i = (*gDemuxMpeg.mInput->read) (gDemuxMpeg.mDummySpace, nLen);
  } else {
    xprintf (VERBOSE|DEMUX, ", unknown (nID = %d)",nID);
    i = (*gDemuxMpeg.mInput->read) (gDemuxMpeg.mDummySpace, nLen);
  }

  xprintf (VERBOSE|DEMUX, ")\n");
}

uint32_t parse_pack()
{
  uint32_t buf ;
  char scratch[1024];
  int mpeg_version;

  xprintf (VERBOSE|DEMUX, "pack {\n");

  /* system_clock_reference */
  buf = read_bytes (1);
  xprintf (VERBOSE|DEMUX|VIDEO, "  mpeg version : %02x",buf>>4);

  if ((buf>>4) == 4) {
     xprintf (VERBOSE|DEMUX|VIDEO, "  => mpeg II \n");
     buf = read_bytes(2);
     mpeg_version = 2;
  } else {
     xprintf (VERBOSE|DEMUX|VIDEO, "  => mpeg I \n");
     mpeg_version = 1;
  }

  buf = read_bytes (2);
  buf = read_bytes (2);

  /* mux_rate */

  buf = read_bytes (3) ;

  /* printf ("  mux_rate = %06x\n",buf); */

  /* system header */

  buf = read_bytes (4) ;

  /* printf ("  code = %08x\n",buf);*/

  if (buf == 0x000001bb) {
    buf = read_bytes (2);
    xprintf (VERBOSE|DEMUX, "  system_header (%d +6 bytes)\n",buf);

    (*gDemuxMpeg.mInput->read) (scratch,buf);

    buf = read_bytes (4) ;
  }
  
  /* printf ("  code = %08x\n",buf); */

  while ( ((buf & 0xFFFFFF00) == 0x00000100)
	  && ((buf & 0xff) != 0xba) ) {

    if (gDemuxMpeg.mnStatus != DEMUX_OK)
      return buf;

    if (mpeg_version == 1)
      parse_mpeg1_packet (buf & 0xFF);
    else
      parse_mpeg2_packet (buf & 0xFF);

    buf = read_bytes (4);
    xprintf (VERBOSE|DEMUX, "  code = %08x\n",buf); 
  }
  
  xprintf (VERBOSE|DEMUX, "}\n");

  return buf;

}

void demux_mpeg_resync (uint32_t buf) {

  while ((buf !=0x000001ba) && (gDemuxMpeg.mnStatus == DEMUX_OK)) {
    xprintf (VERBOSE|DEMUX, "resync : %08x\n",buf);
    buf = (buf << 8) | read_bytes (1);
  }
}

void *demux_mpeg_loop (void *dummy) {
  uint32_t buf;

  do {
    buf = parse_pack ();

    if (buf != 0x000001ba)
      demux_mpeg_resync (buf);
    
  } while (gDemuxMpeg.mnStatus == DEMUX_OK) ;

  xprintf (VERBOSE|DEMUX, "demux loop finished (status: %d, buf:%x)\n",
	   gDemuxMpeg.mnStatus, buf);

  return NULL;
}

void demux_mpeg_stop () {
  void *p;

  gDemuxMpeg.mnStatus = DEMUX_FINISHED;

  fifo_buffer_clear(gDemuxMpeg.mBufVideo);
  fifo_buffer_clear(gDemuxMpeg.mBufAudio);

  pthread_join (gDemuxMpeg.mThread, &p);
}

int demux_mpeg_get_status () {
  return gDemuxMpeg.mnStatus;
}

void demux_mpeg_prefill () {

  uint32_t buf;

  while ((gDemuxMpeg.mnStatus == DEMUX_OK) 
	 && !buffer_pool_isempty ()) {

    buf = parse_pack ();

    if (buf != 0x000001ba)
      demux_mpeg_resync (buf);
  }
}

void demux_mpeg_start (input_plugin_t *input_plugin,
		       fifo_buffer_t *bufVideo, fifo_buffer_t *bufAudio,
		       off_t pos) 
{
  gDemuxMpeg.mInput       = input_plugin;
  gDemuxMpeg.mBufVideo    = bufVideo;
  gDemuxMpeg.mBufAudio    = bufAudio;

  gDemuxMpeg.mnStatus     = DEMUX_OK;

  if ((gDemuxMpeg.mInput->get_capabilities () & INPUT_CAP_SEEKABLE) != 0 ) {
    xprintf (VERBOSE|DEMUX, "=>seek to %Ld\n",pos);

    gDemuxMpeg.mInput->seek (pos+4, SEEK_SET);
  }

  demux_mpeg_prefill ();

  pthread_create (&gDemuxMpeg.mThread, NULL, demux_mpeg_loop, NULL) ;
}

void demux_mpeg_select_audio_channel (int nChannel) {
  gDemuxMpeg.mnAudioChannel = nChannel;
}

int demux_mpeg_can_handle (char *MRL) {
  char *media;
  char *ending;
  
  /*
   * check media type
   */
  
  media = strstr(MRL, "://");
  if(media) {
    if((!(strncasecmp(MRL, "stdin", 5))) || (!(strncasecmp(MRL, "fifo", 4)))) {
      if(!(strncasecmp((media+3), "mpeg1", 5))) {
	return 1;
      }
      else if(!(strncasecmp((media+3), "mpeg2", 5))) {
	return 0;
      }
      fprintf(stderr, "You should specify mpeg(mpeg1/mpeg2) stream type.\n");
      return 0;
    }
    else if(strncasecmp(MRL, "file", 4)) {
      return 0;
    }
  }

  /*
   * check ending
   */

  ending = strrchr(MRL, '.');
  xprintf(VERBOSE|DEMUX, "demux_mpeg_can_handle: ending %s of %s\n", ending, MRL);
  
  if(!ending)
    return 0;
  
  if(strcasecmp(ending, ".mpg") 
     && (strcasecmp(ending, ".mpeg"))) {
    return 0;
  }

  return 1;
}

static demux_functions_t demux_mpeg_functions = {
  demux_mpeg_can_handle,
  demux_mpeg_start,
  demux_mpeg_stop,
  demux_mpeg_get_status,
  demux_mpeg_select_audio_channel
};

demux_functions_t *init_demux_mpeg() {
  return &demux_mpeg_functions;
}
